from abc import ABC, abstractmethod
from typing import List, Optional, Union

import pandas
from jinja2 import Template


class QueueAdapterAbstractClass(ABC):
    @abstractmethod
    def submit_job(
        self,
        queue: Optional[str] = None,
        job_name: Optional[str] = None,
        working_directory: Optional[str] = None,
        cores: Optional[int] = None,
        memory_max: Optional[int] = None,
        run_time_max: Optional[int] = None,
        dependency_list: Optional[List[str]] = None,
        command: Optional[str] = None,
        submission_template: Optional[Union[str, Template]] = None,
        **kwargs,
    ) -> Union[int, None]:
        pass

    @abstractmethod
    def enable_reservation(self, process_id: int):
        pass

    @abstractmethod
    def delete_job(self, process_id: int) -> Union[str, None]:
        pass

    @abstractmethod
    def get_queue_status(self, user: Optional[str] = None) -> pandas.DataFrame:
        """
        Get the status of the queue.

        Args:
            user (str): The user to filter the queue status for.

        Returns:
            pandas.DataFrame: The queue status.
        """
        pass

    @abstractmethod
    def get_status_of_my_jobs(self) -> pandas.DataFrame:
        """
        Get the status of the user's jobs.

        Returns:
            pandas.DataFrame: The status of the user's jobs.
        """
        pass

    @abstractmethod
    def get_status_of_job(self, process_id: int) -> Union[str, None]:
        """
        Get the status of a job.

        Args:
            process_id (int): The process ID.

        Returns:
            str: The status of the job.results_lst.append(df_selected.values[0])
        """
        pass

    @abstractmethod
    def get_status_of_jobs(self, process_id_lst: List[int]) -> List[str]:
        """
        Get the status of multiple jobs.

        Args:
            process_id_lst (list[int]): List of process IDs.

        Returns:
            list[str]: List of job statuses.
        """
        pass



import os
from typing import List, Optional, Union

from jinja2 import Template

from pysqa.wrapper.abstract import SchedulerCommands

template = """\
#!/bin/bash
#MSUB -N {{job_name}}
{%- if memory_max %}
#MSUB -l pmem={{ memory_max| int }}gb
{%- endif %}
{%- if run_time_max %}
#$ -l walltime={{run_time_max}}
{%- endif %}

{{command}}
"""


class MoabCommands(SchedulerCommands):
    @property
    def submit_job_command(self) -> list[str]:
        """
        Get the command to submit a job.

        Returns:
            list[str]: The command to submit a job.
        """
        return ["msub"]

    @property
    def delete_job_command(self) -> list[str]:
        """
        Get the command to delete a job.

        Returns:
            list[str]: The command to delete a job.
        """
        return ["mjobctl", "-c"]

    @property
    def get_queue_status_command(self) -> list[str]:
        """
        Get the command to get the queue status.

        Returns:
            list[str]: The command to get the queue status.
        """
        return ["mdiag", "-x"]

    def render_submission_template(
        self,
        command: str,
        job_name: str = "pysqa",
        working_directory: str = os.path.abspath("."),
        cores: int = 1,
        memory_max: Optional[int] = None,
        run_time_max: Optional[int] = None,
        dependency_list: Optional[List[int]] = None,
        submission_template: Union[str, Template] = template,
        **kwargs,
    ) -> str:
        """
        Generate the job submission template.

        Args:
            command (str, optional): The command to be executed.
            job_name (str, optional): The job name. Defaults to "pysqa".
            working_directory (str, optional): The working directory. Defaults to ".".
            cores (int, optional): The number of cores. Defaults to 1.
            memory_max (int, optional): The maximum memory. Defaults to None.
            run_time_max (int, optional): The maximum run time. Defaults to None.
            dependency_list (list[int], optional): The list of dependency job IDs. Defaults to None.
            submission_template (str): Submission script template pysqa.wrapper.flux.template

        Returns:
            str: The rendered job submission template.
        """
        return super().render_submission_template(
            command=command,
            job_name=job_name,
            working_directory=working_directory,
            cores=cores,
            memory_max=memory_max,
            run_time_max=run_time_max,
            dependency_list=dependency_list,
            submission_template=submission_template,
            **kwargs,
        )



import os
from typing import List, Optional, Union

import defusedxml.ElementTree as ETree
import pandas
from jinja2 import Template

from pysqa.wrapper.abstract import SchedulerCommands

template = """\
#!/bin/bash
#$ -N {{job_name}}
#$ -wd {{working_directory}}
{%- if cores %}
#$ -pe {{partition}} {{cores}}
{%- endif %}
{%- if memory_max %}
#$ -l h_vmem={{memory_max}}
{%- endif %}
{%- if run_time_max %}
#$ -l h_rt={{run_time_max}}
{%- endif %}
#$ -o time.out
#$ -e error.out

{{command}}
"""


class SunGridEngineCommands(SchedulerCommands):
    @property
    def submit_job_command(self) -> list[str]:
        """Return the command to submit a job."""
        return ["qsub", "-terse"]

    @property
    def delete_job_command(self) -> list[str]:
        """Return the command to delete a job."""
        return ["qdel"]

    @property
    def enable_reservation_command(self) -> list[str]:
        """Return the command to enable job reservation."""
        return ["qalter", "-R", "y"]

    @property
    def get_queue_status_command(self) -> list[str]:
        """Return the command to get the queue status."""
        return ["qstat", "-xml"]

    @staticmethod
    def convert_queue_status(queue_status_output: str) -> pandas.DataFrame:
        """Convert the queue status output to a pandas DataFrame.

        Args:
            queue_status_output: The output of the queue status command.

        Returns:
            A pandas DataFrame containing the converted queue status.

        """

        def leaf_to_dict(leaf):
            return [
                {sub_child.tag: sub_child.text for sub_child in child} for child in leaf
            ]

        tree = ETree.fromstring(queue_status_output)
        df_running_jobs = pandas.DataFrame(leaf_to_dict(leaf=tree[0]))
        df_pending_jobs = pandas.DataFrame(leaf_to_dict(leaf=tree[1]))
        df_merge = pandas.concat([df_running_jobs, df_pending_jobs], sort=True)
        df_merge.loc[df_merge.state == "r", "state"] = "running"
        df_merge.loc[df_merge.state == "qw", "state"] = "pending"
        df_merge.loc[df_merge.state == "Eqw", "state"] = "error"
        return pandas.DataFrame(
            {
                "jobid": pandas.to_numeric(df_merge.JB_job_number),
                "user": df_merge.JB_owner,
                "jobname": df_merge.JB_name,
                "status": df_merge.state,
                "working_directory": [""] * len(df_merge),
            }
        )

    def render_submission_template(
        self,
        command: str,
        job_name: str = "pysqa",
        working_directory: str = os.path.abspath("."),
        cores: int = 1,
        memory_max: Optional[int] = None,
        run_time_max: Optional[int] = None,
        dependency_list: Optional[List[int]] = None,
        submission_template: Union[str, Template] = template,
        **kwargs,
    ) -> str:
        """
        Generate the job submission template.

        Args:
            command (str, optional): The command to be executed.
            job_name (str, optional): The job name. Defaults to "pysqa".
            working_directory (str, optional): The working directory. Defaults to ".".
            cores (int, optional): The number of cores. Defaults to 1.
            memory_max (int, optional): The maximum memory. Defaults to None.
            run_time_max (int, optional): The maximum run time. Defaults to None.
            dependency_list (list[int], optional): The list of dependency job IDs. Defaults to None.
            submission_template (str): Submission script template pysqa.wrapper.flux.template

        Returns:
            str: The rendered job submission template.
        """
        return super().render_submission_template(
            command=command,
            job_name=job_name,
            working_directory=working_directory,
            cores=cores,
            memory_max=memory_max,
            run_time_max=run_time_max,
            dependency_list=dependency_list,
            submission_template=submission_template,
            **kwargs,
        )



import re
from typing import Optional, Tuple, Union


def check_queue_parameters(
    active_queue: Optional[dict] = None,
    cores: int = 1,
    run_time_max: Optional[int] = None,
    memory_max: Optional[int] = None,
) -> Tuple[Union[float, int, None], Union[float, int, None], Union[float, int, None]]:
    """
    Check the parameters of a queue.

    Args:

        cores (int, optional): The number of cores. Defaults to 1.
        run_time_max (int, optional): The maximum run time. Defaults to None.
        memory_max (int, optional): The maximum memory. Defaults to None.
        active_queue (dict, optional): The active queue. Defaults to None.

    Returns:
        list: A list of queue parameters [cores, run_time_max, memory_max].
    """
    cores = value_in_range(
        value=cores,
        value_min=active_queue["cores_min"],
        value_max=active_queue["cores_max"],
    )
    run_time_max = value_in_range(
        value=run_time_max, value_max=active_queue["run_time_max"]
    )
    memory_max = value_in_range(value=memory_max, value_max=active_queue["memory_max"])
    return cores, run_time_max, memory_max


def value_error_if_none(value: str) -> None:
    """
    Raise a ValueError if the value is None or not a string.

    Args:
        value (str/None): The value to check.

    Raises:
        ValueError: If the value is None.
        TypeError: If the value is not a string.
    """
    if value is None:
        raise ValueError("Value cannot be None.")
    if not isinstance(value, str):
        raise TypeError()


def value_in_range(
    value: Union[int, float, None],
    value_min: Union[int, float, None] = None,
    value_max: Union[int, float, None] = None,
) -> Union[int, float, None]:
    """
    Check if a value is within a specified range.

    Args:
        value (int/float/None): The value to check.
        value_min (int/float/None): The minimum value. Defaults to None.
        value_max (int/float/None): The maximum value. Defaults to None.

    Returns:
        int/float/None: The value if it is within the range, otherwise the minimum or maximum value.
    """

    if value is not None:
        value_, value_min_, value_max_ = [
            (
                _memory_spec_string_to_value(v)
                if v is not None and isinstance(v, str)
                else v
            )
            for v in (value, value_min, value_max)
        ]
        # ATTENTION: '60000' is interpreted as '60000M' since default magnitude is 'M'
        # ATTENTION: int('60000') is interpreted as '60000B' since _memory_spec_string_to_value return the size in
        # ATTENTION: bytes, as target_magnitude = 'b'
        # We want to compare the the actual (k,m,g)byte value if there is any
        if value_min_ is not None and value_ < value_min_:
            return value_min
        if value_max_ is not None and value_ > value_max_:
            return value_max
        return value
    else:
        if value_min is not None:
            return value_min
        if value_max is not None:
            return value_max
        return value


def _is_memory_string(value: str) -> bool:
    """
    Check if a string specifies a certain amount of memory.

    Args:
        value (str): The string to check.

    Returns:
        bool: True if the string matches a memory specification, False otherwise.
    """
    memory_spec_pattern = r"[0-9]+[bBkKmMgGtT]?"
    return re.findall(memory_spec_pattern, value)[0] == value


def _memory_spec_string_to_value(
    value: Union[str, int, float],
    default_magnitude: str = "m",
    target_magnitude: str = "b",
) -> Union[int, float]:
    """
    Converts a valid memory string (tested by _is_memory_string) into an integer/float value of desired
    magnitude `default_magnitude`. If it is a plain integer string (e.g.: '50000') it will be interpreted with
    the magnitude passed in by the `default_magnitude`. The output will rescaled to `target_magnitude`

    Args:
        value (str): The string to convert.
        default_magnitude (str): The magnitude for interpreting plain integer strings [b, B, k, K, m, M, g, G, t, T]. Defaults to "m".
        target_magnitude (str): The magnitude to which the output value should be converted [b, B, k, K, m, M, g, G, t, T]. Defaults to "b".

    Returns:
        Union[int, float]: The value of the string in `target_magnitude` units.
    """
    magnitude_mapping = {"b": 0, "k": 1, "m": 2, "g": 3, "t": 4}
    if _is_memory_string(value):
        integer_pattern = r"[0-9]+"
        magnitude_pattern = r"[bBkKmMgGtT]+"
        integer_value = int(re.findall(integer_pattern, value)[0])

        magnitude = re.findall(magnitude_pattern, value)
        if len(magnitude) > 0:
            magnitude = magnitude[0].lower()
        else:
            magnitude = default_magnitude.lower()
        # Convert it to default magnitude = megabytes
        return (integer_value * 1024 ** magnitude_mapping[magnitude]) / (
            1024 ** magnitude_mapping[target_magnitude]
        )
    else:
        return value



import getpass
import json
import os
import warnings
from typing import List, Optional, Union

import pandas
import paramiko
from tqdm import tqdm

from pysqa.base.config import QueueAdapterWithConfig
from pysqa.base.core import execute_command


class RemoteQueueAdapter(QueueAdapterWithConfig):
    """
    A class representing a remote queue adapter.

    Args:
        config (dict): The configuration dictionary.
        directory (str, optional): The directory path. Defaults to "~/.queues".
        execute_command (callable, optional): The execute command function. Defaults to execute_command.

    Attributes:
        _ssh_host (str): The SSH host.
        _ssh_username (str): The SSH username.
        _ssh_known_hosts (str): The path to the known hosts file.
        _ssh_key (str): The path to the SSH key file.
        _ssh_ask_for_password (bool): Flag indicating whether to ask for SSH password.
        _ssh_password (str): The SSH password.
        _ssh_key_passphrase (str): The SSH key passphrase.
        _ssh_two_factor_authentication (bool): Flag indicating whether to use two-factor authentication.
        _ssh_authenticator_service (str): The SSH authenticator service.
        _ssh_proxy_host (str): The SSH proxy host.
        _ssh_remote_config_dir (str): The remote configuration directory.
        _ssh_remote_path (str): The remote path.
        _ssh_local_path (str): The local path.
        _ssh_delete_file_on_remote (bool): Flag indicating whether to delete files on the remote host.
        _ssh_port (int): The SSH port.
        _ssh_continous_connection (bool): Flag indicating whether to use continuous SSH connection.
        _ssh_connection (None or paramiko.SSHClient): The SSH connection object.
        _ssh_proxy_connection (None or paramiko.SSHClient): The SSH proxy connection object.
        _remote_flag (bool): Flag indicating whether the adapter is for remote queue.

    Methods:
        convert_path_to_remote(path: str) -> str:
            Converts a local path to a remote path.

        submit_job(queue: Optional[str] = None, job_name: Optional[str] = None, working_directory: Optional[str] = None,
                   cores: Optional[int] = None, memory_max: Optional[int] = None, run_time_max: Optional[int] = None,
                   dependency_list: Optional[list[str]] = None, command: Optional[str] = None, **kwargs) -> int:
            Submits a job to the remote queue.

        enable_reservation(process_id: int) -> str:
            Enables a reservation for a job.

        delete_job(process_id: int) -> str:
            Deletes a job from the remote queue.

        get_queue_status(user: Optional[str] = None) -> pandas.DataFrame:
            Retrieves the queue status.

        get_job_from_remote(working_directory: str):
            Retrieves the results of a calculation executed on a remote host.

        transfer_file(file: str, transfer_back: bool = False, delete_file_on_remote: bool = False):
            Transfers a file to/from the remote host.

        __del__():
            Closes the SSH connections.

        _check_ssh_connection():
            Checks if an SSH connection is open.

        _transfer_files(file_dict: dict, sftp=None, transfer_back: bool = False):
            Transfers files to/from the remote host.

        _open_ssh_connection() -> paramiko.SSHClient:
            Opens an SSH connection.
    """

    def __init__(
        self,
        config: dict,
        directory: str = "~/.queues",
        execute_command: callable = execute_command,
    ):
        super(RemoteQueueAdapter, self).__init__(
            config=config, directory=directory, execute_command=execute_command
        )
        self._ssh_host = config["ssh_host"]
        self._ssh_username = config["ssh_username"]
        self._ssh_known_hosts = os.path.abspath(
            os.path.expanduser(config["known_hosts"])
        )
        if "ssh_key" in config.keys():
            self._ssh_key = os.path.abspath(os.path.expanduser(config["ssh_key"]))
            self._ssh_ask_for_password = False
        else:
            self._ssh_key = None
        if "ssh_password" in config.keys():
            self._ssh_password = config["ssh_password"]
            self._ssh_ask_for_password = False
        else:
            self._ssh_password = None
        if "ssh_ask_for_password" in config.keys():
            self._ssh_ask_for_password = config["ssh_ask_for_password"]
        else:
            self._ssh_ask_for_password = False
        if "ssh_key_passphrase" in config.keys():
            self._ssh_key_passphrase = config["ssh_key_passphrase"]
        else:
            self._ssh_key_passphrase = None
        if "ssh_two_factor_authentication" in config.keys():
            self._ssh_two_factor_authentication = config[
                "ssh_two_factor_authentication"
            ]
        else:
            self._ssh_two_factor_authentication = False
        if "ssh_authenticator_service" in config.keys():
            self._ssh_authenticator_service = config["ssh_authenticator_service"]
            self._ssh_two_factor_authentication = True
        else:
            self._ssh_authenticator_service = None
        if "ssh_proxy_host" in config.keys():
            self._ssh_proxy_host = config["ssh_proxy_host"]
        else:
            self._ssh_proxy_host = None
        self._ssh_remote_config_dir = config["ssh_remote_config_dir"]
        self._ssh_remote_path = config["ssh_remote_path"]
        self._ssh_local_path = os.path.abspath(
            os.path.expanduser(config["ssh_local_path"])
        )
        if "ssh_delete_file_on_remote" in config.keys():
            self._ssh_delete_file_on_remote = config["ssh_delete_file_on_remote"]
        else:
            self._ssh_delete_file_on_remote = True
        if "ssh_port" in config.keys():
            self._ssh_port = config["ssh_port"]
        else:
            self._ssh_port = 22
        if "ssh_continous_connection" in config.keys():
            self._ssh_continous_connection = config["ssh_continous_connection"]
        else:
            self._ssh_continous_connection = False
        self._ssh_connection = None
        self._ssh_proxy_connection = None
        self._remote_flag = True

    def convert_path_to_remote(self, path: str) -> str:
        """
        Converts a local path to a remote path.

        Args:
            path (str): The local path.

        Returns:
            str: The remote path.
        """
        working_directory = os.path.abspath(os.path.expanduser(path))
        return self._get_remote_working_dir(working_directory=working_directory)

    def submit_job(
        self,
        queue: Optional[str] = None,
        job_name: Optional[str] = None,
        working_directory: Optional[str] = None,
        cores: Optional[int] = None,
        memory_max: Optional[int] = None,
        run_time_max: Optional[int] = None,
        dependency_list: Optional[list[str]] = None,
        command: Optional[str] = None,
        **kwargs,
    ) -> int:
        """
        Submits a job to the remote queue.

        Args:
            queue (str, optional): The queue name.
            job_name (str, optional): The job name.
            working_directory (str, optional): The working directory.
            cores (int, optional): The number of cores.
            memory_max (int, optional): The maximum memory.
            run_time_max (int, optional): The maximum run time.
            dependency_list (list[str], optional): The list of job dependencies.
            command (str, optional): The command to execute.

        Returns:
            int: The process ID of the submitted job.
        """
        if dependency_list is not None:
            raise NotImplementedError(
                "Submitting jobs with dependencies to a remote cluster is not yet supported."
            )
        self._transfer_data_to_remote(working_directory=working_directory)
        output = self._execute_remote_command(command=command)
        return int(output.split()[-1])

    def enable_reservation(self, process_id: int) -> str:
        """
        Enables a reservation for a job.

        Args:
            process_id (int): The process ID.

        Returns:
            str: The output of the reservation command.
        """
        return self._execute_remote_command(
            command=self._reservation_command(job_id=process_id)
        )

    def delete_job(self, process_id: int) -> str:
        """
        Deletes a job from the remote queue.

        Args:
            process_id (int): The process ID.

        Returns:
            str: The output of the delete command.
        """
        return self._execute_remote_command(
            command=self._delete_command(job_id=process_id)
        )

    def get_queue_status(self, user: Optional[str] = None) -> pandas.DataFrame:
        """
        Retrieves the queue status.

        Args:
            user (str, optional): The username.

        Returns:
            pandas.DataFrame: The queue status.
        """
        df = pandas.DataFrame(
            json.loads(
                self._execute_remote_command(command=self._get_queue_status_command())
            )
        )
        if user is None:
            return df
        else:
            return df[df["user"] == user]

    def get_job_from_remote(self, working_directory: str):
        """
        Retrieves the results of a calculation executed on a remote host.

        Args:
            working_directory (str): The local working directory.
        """
        working_directory = os.path.abspath(os.path.expanduser(working_directory))
        remote_working_directory = self._get_remote_working_dir(
            working_directory=working_directory
        )
        remote_dict = json.loads(
            self._execute_remote_command(
                command="python -m pysqa --list --working_directory "
                + remote_working_directory
            )
        )
        for d in remote_dict["dirs"]:
            local_dir = self._get_file_transfer(
                file=d, local_dir=remote_working_directory, remote_dir=working_directory
            )
            os.makedirs(local_dir, exist_ok=True)
        file_dict = {}
        for f in remote_dict["files"]:
            local_file = self._get_file_transfer(
                file=f, local_dir=remote_working_directory, remote_dir=working_directory
            )
            file_dict[local_file] = f
        self._transfer_files(file_dict=file_dict, sftp=None, transfer_back=True)
        if self._ssh_delete_file_on_remote:
            self._execute_remote_command(command="rm -r " + remote_working_directory)

    def transfer_file(
        self,
        file: str,
        transfer_back: bool = False,
        delete_file_on_remote: bool = False,
    ):
        """
        Transfers a file to/from the remote host.

        Args:
            file (str): The file path.
            transfer_back (bool, optional): Flag indicating whether to transfer the file back to the local host.
            delete_file_on_remote (bool, optional): Flag indicating whether to delete the file on the remote host.
        """
        working_directory = os.path.abspath(os.path.expanduser(file))
        remote_working_directory = self._get_remote_working_dir(
            working_directory=working_directory
        )
        self._create_remote_dir(directory=os.path.dirname(remote_working_directory))
        self._transfer_files(
            file_dict={working_directory: remote_working_directory},
            sftp=None,
            transfer_back=transfer_back,
        )
        if self._ssh_delete_file_on_remote and transfer_back and delete_file_on_remote:
            self._execute_remote_command(command="rm " + remote_working_directory)

    def __del__(self):
        """
        Closes the SSH connections.
        """
        if self._ssh_connection is not None:
            self._ssh_connection.close()
        if self._ssh_proxy_connection is not None:
            self._ssh_proxy_connection.close()

    def _check_ssh_connection(self):
        """
        Checks if an SSH connection is open.
        """
        if self._ssh_connection is None:
            self._ssh_connection = self._open_ssh_connection()

    def _transfer_files(self, file_dict: dict, sftp=None, transfer_back: bool = False):
        """
        Transfers files to/from the remote host.

        Args:
            file_dict (dict): The dictionary containing the file paths.
            sftp (None or paramiko.SFTPClient, optional): The SFTP client object.
            transfer_back (bool, optional): Flag indicating whether to transfer the files back to the local host.
        """
        if sftp is None:
            if self._ssh_continous_connection:
                self._check_ssh_connection()
                ssh = self._ssh_connection
            else:
                ssh = self._open_ssh_connection()
            sftp_client = ssh.open_sftp()
        else:
            sftp_client = sftp
        for file_src, file_dst in tqdm(file_dict.items()):
            if transfer_back:
                try:
                    # Check remote file existence.
                    # If the remote file does not exist, sftp_client.get() will make the local file empty
                    # sftp_client.stat() can throw an exception early to prevent the execution of sftp_client.get().
                    sftp_client.stat(file_dst)
                    sftp_client.get(file_dst, file_src)
                except FileNotFoundError:
                    pass
            else:
                sftp_client.put(file_src, file_dst)
        if sftp is None:
            sftp_client.close()

    def _open_ssh_connection(self) -> paramiko.SSHClient:
        """
        Opens an SSH connection.

        Returns:
            paramiko.SSHClient: The SSH connection object.
        """
        ssh = paramiko.SSHClient()
        ssh.load_host_keys(self._ssh_known_hosts)
        if (
            self._ssh_key is not None
            and self._ssh_key_passphrase is not None
            and not self._ssh_ask_for_password
        ):
            ssh.connect(
                hostname=self._ssh_host,
                port=self._ssh_port,
                username=self._ssh_username,
                key_filename=self._ssh_key,
                passphrase=self._ssh_key_passphrase,
            )
        elif self._ssh_key is not None and not self._ssh_ask_for_password:
            ssh.connect(
                hostname=self._ssh_host,
                port=self._ssh_port,
                username=self._ssh_username,
                key_filename=self._ssh_key,
            )
        elif (
            self._ssh_password is not None
            and self._ssh_authenticator_service is None
            and not self._ssh_two_factor_authentication
            and not self._ssh_ask_for_password
        ):
            ssh.connect(
                hostname=self._ssh_host,
                port=self._ssh_port,
                username=self._ssh_username,
                password=self._ssh_password,
            )
        elif self._ssh_ask_for_password and not self._ssh_two_factor_authentication:
            ssh.connect(
                hostname=self._ssh_host,
                port=self._ssh_port,
                username=self._ssh_username,
                password=getpass.getpass(prompt="SSH Password: ", stream=None),
            )
        elif (
            self._ssh_password is not None
            and self._ssh_authenticator_service is not None
            and self._ssh_two_factor_authentication
        ):

            def authentication(title, instructions, prompt_list):
                from pyauthenticator import get_two_factor_code

                if len(prompt_list) > 0:
                    return [
                        get_two_factor_code(service=self._ssh_authenticator_service)
                    ]
                else:
                    return []

            ssh.connect(
                hostname=self._ssh_host,
                port=self._ssh_port,
                username=self._ssh_username,
                password=self._ssh_password,
            )

            ssh._transport.auth_interactive(
                username=self._ssh_username, handler=authentication, submethods=""
            )
        elif (
            self._ssh_password is not None
            and self._ssh_authenticator_service is None
            and self._ssh_two_factor_authentication
        ):
            ssh.connect(
                hostname=self._ssh_host,
                port=self._ssh_port,
                username=self._ssh_username,
                password=self._ssh_password,
            )
            ssh._transport.auth_interactive_dumb(
                username=self._ssh_username, handler=None, submethods=""
            )
        elif self._ssh_ask_for_password and self._ssh_two_factor_authentication:
            ssh.connect(
                hostname=self._ssh_host,
                port=self._ssh_port,
                username=self._ssh_username,
                password=getpass.getpass(prompt="SSH Password: ", stream=None),
            )
            ssh._transport.auth_interactive_dumb(
                username=self._ssh_username, handler=None, submethods=""
            )
        else:
            raise ValueError("Un-supported authentication method.")

        if self._ssh_proxy_host is not None:
            client_new = paramiko.SSHClient()
            client_new.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            vmtransport = ssh.get_transport()
            vmchannel = vmtransport.open_channel(
                kind="direct-tcpip",
                dest_addr=(self._ssh_proxy_host, self._ssh_port),
                src_addr=(self._ssh_host, self._ssh_port),
            )
            client_new.connect(
                hostname=self._ssh_proxy_host,
                username=self._ssh_username,
                sock=vmchannel,
            )
            self._ssh_proxy_connection = ssh
            return client_new
        else:
            return ssh

    def _remote_command(self) -> str:
        """
        Returns the command to execute on the remote host.

        Returns:
            str: The remote command.
        """
        return "python -m pysqa --config_directory " + self._ssh_remote_config_dir + " "

    def _get_queue_status_command(self) -> str:
        """
        Returns the command to get the queue status on the remote host.

        Returns:
            str: The queue status command.
        """
        return self._remote_command() + "--status"

    def _submit_command(
        self,
        queue: Optional[str] = None,
        job_name: Optional[str] = None,
        working_directory: Optional[str] = None,
        cores: Optional[int] = None,
        memory_max: Optional[int] = None,
        run_time_max: Optional[int] = None,
        command_str: Optional[str] = None,
    ) -> str:
        """
        Generates the command to submit a job to the remote host.

        Args:
            queue (str, optional): The queue to submit the job to.
            job_name (str, optional): The name of the job.
            working_directory (str, optional): The working directory for the job.
            cores (int, optional): The number of cores required for the job.
            memory_max (int, optional): The maximum memory required for the job.
            run_time_max (int, optional): The maximum run time for the job.
            command_str (str, optional): The command to be executed by the job.

        Returns:
            str: The submit command.
        """
        command = self._remote_command() + "--submit "
        if queue is not None:
            command += "--queue " + queue + " "
        if job_name is not None:
            command += "--job_name " + job_name + " "
        if working_directory is not None:
            command += "--working_directory " + working_directory + " "
        if cores is not None:
            command += "--cores " + str(cores) + " "
        if memory_max is not None:
            command += "--memory " + str(memory_max) + " "
        if run_time_max is not None:
            command += "--run_time " + str(run_time_max) + " "
        if command_str is not None:
            command += '--command "' + command_str + '" '
        return command

    def _delete_command(self, job_id: int) -> str:
        """
        Generates the command to delete a job on the remote host.

        Args:
            job_id (int): The ID of the job to delete.

        Returns:
            str: The delete command.
        """
        return self._remote_command() + "--delete --id " + str(job_id)

    def _reservation_command(self, job_id: int) -> str:
        """
        Generates the command to reserve a job on the remote host.

        Args:
            job_id (int): The ID of the job to reserve.

        Returns:
            str: The reservation command.
        """
        return self._remote_command() + "--reservation --id " + str(job_id)

    def _execute_remote_command(self, command: str) -> str:
        """
        Executes a remote command on the SSH connection.

        Args:
            command (str): The command to execute.

        Returns:
            str: The output of the command.
        """
        if self._ssh_continous_connection:
            self._check_ssh_connection()
            ssh = self._ssh_connection
        else:
            ssh = self._open_ssh_connection()
        stdin, stdout, stderr = ssh.exec_command(command)
        warnings.warn(stderr.read().decode())
        output = stdout.read().decode()
        if not self._ssh_continous_connection:
            ssh.close()
        return output

    def _get_remote_working_dir(self, working_directory: str) -> str:
        """
        Get the remote working directory path.

        Args:
            working_directory (str): The local working directory path.

        Returns:
            str: The remote working directory path.
        """
        return os.path.join(
            self._ssh_remote_path,
            os.path.relpath(working_directory, self._ssh_local_path),
        )

    def _create_remote_dir(self, directory: Union[str, List[str]]) -> None:
        """
        Creates a remote directory on the SSH server.

        Args:
            directory (Union[str, List[str]]): The directory or list of directories to create.

        Raises:
            TypeError: If the directory argument is not a string or a list.
        """
        if isinstance(directory, str):
            self._execute_remote_command(command="mkdir -p " + directory)
        elif isinstance(directory, list):
            command = "mkdir -p "
            for d in directory:
                command += d + " "
            self._execute_remote_command(command=command)
        else:
            raise TypeError("Invalid directory argument type.")

    def _transfer_data_to_remote(self, working_directory: str) -> None:
        """
        Transfers data from the local machine to the remote host.

        Args:
            working_directory (str): The local working directory path.

        Returns:
            None
        """
        working_directory = os.path.abspath(os.path.expanduser(working_directory))
        remote_working_directory = self._get_remote_working_dir(
            working_directory=working_directory
        )
        file_dict = {}
        new_dir_list = []
        for p, folder, files in os.walk(working_directory):
            new_dir_list.append(
                self._get_file_transfer(
                    file=p,
                    local_dir=working_directory,
                    remote_dir=remote_working_directory,
                )
            )
            for f in files:
                file_path = os.path.join(p, f)
                file_dict[file_path] = self._get_file_transfer(
                    file=file_path,
                    local_dir=working_directory,
                    remote_dir=remote_working_directory,
                )
        self._create_remote_dir(directory=new_dir_list)
        self._transfer_files(file_dict=file_dict, sftp=None, transfer_back=False)

    def _get_user(self) -> str:
        """
        Get the username used for SSH connection.

        Returns:
            str: The username.
        """
        return self._ssh_username

    @staticmethod
    def _get_file_transfer(file: str, local_dir: str, remote_dir: str) -> str:
        """
        Get the absolute path of the file on the remote host.

        Args:
            file (str): The file path.
            local_dir (str): The local working directory path.
            remote_dir (str): The remote working directory path.

        Returns:
            str: The absolute path of the file on the remote host.
        """
        return os.path.abspath(
            os.path.join(remote_dir, os.path.relpath(file, local_dir))
        )



import os
from typing import List, Optional, Union

import pandas
from jinja2 import Template

from pysqa.wrapper.abstract import SchedulerCommands

template = """\
#!/bin/bash
#SBATCH --output=time.out
#SBATCH --job-name={{job_name}}
#SBATCH --chdir={{working_directory}}
#SBATCH --get-user-env=L
#SBATCH --partition={{partition}}
{%- if run_time_max %}
#SBATCH --time={{ [1, run_time_max // 60]|max }}
{%- endif %}
{%- if dependency %}
#SBATCH --dependency=afterok:{{ dependency | join(',') }}
{%- endif %}
{%- if memory_max %}
#SBATCH --mem={{memory_max}}G
{%- endif %}
#SBATCH --cpus-per-task={{cores}}

{{command}}
"""


class SlurmCommands(SchedulerCommands):
    @property
    def submit_job_command(self) -> list[str]:
        """Returns the command to submit a job to Slurm."""
        return ["sbatch", "--parsable"]

    @property
    def delete_job_command(self) -> list[str]:
        """Returns the command to delete a job from Slurm."""
        return ["scancel"]

    @property
    def get_queue_status_command(self) -> list[str]:
        """Returns the command to get the queue status from Slurm."""
        return ["squeue", "--format", "%A|%u|%t|%.15j|%Z", "--noheader"]

    @staticmethod
    def get_job_id_from_output(queue_submit_output: str) -> int:
        """Extracts the job ID from the output of the job submission command."""
        return int(queue_submit_output.splitlines()[-1].rstrip().lstrip().split()[-1])

    @staticmethod
    def convert_queue_status(queue_status_output: str) -> pandas.DataFrame:
        """Converts the queue status output from Slurm into a pandas DataFrame."""
        line_split_lst = [line.split("|") for line in queue_status_output.splitlines()]
        if len(line_split_lst) != 0:
            job_id_lst, user_lst, status_lst, job_name_lst, working_directory_lst = zip(
                *[
                    (int(jobid), user, status.lower(), jobname, working_directory)
                    for jobid, user, status, jobname, working_directory in line_split_lst
                ]
            )
        else:
            job_id_lst, user_lst, status_lst, job_name_lst, working_directory_lst = (
                [],
                [],
                [],
                [],
                [],
            )
        df = pandas.DataFrame(
            {
                "jobid": job_id_lst,
                "user": user_lst,
                "jobname": job_name_lst,
                "status": status_lst,
                "working_directory": working_directory_lst,
            }
        )
        df.loc[df.status == "r", "status"] = "running"
        df.loc[df.status == "pd", "status"] = "pending"
        return df

    @staticmethod
    def dependencies(dependency_list: list[str]) -> list[str]:
        """Returns the dependency options for job submission."""
        if dependency_list is not None:
            return ["--dependency=afterok:" + ",".join(dependency_list)]
        else:
            return []

    def render_submission_template(
        self,
        command: str,
        job_name: str = "pysqa",
        working_directory: str = os.path.abspath("."),
        cores: int = 1,
        memory_max: Optional[int] = None,
        run_time_max: Optional[int] = None,
        dependency_list: Optional[List[int]] = None,
        submission_template: Union[str, Template] = template,
        **kwargs,
    ) -> str:
        """
        Generate the job submission template.

        Args:
            command (str, optional): The command to be executed.
            job_name (str, optional): The job name. Defaults to "pysqa".
            working_directory (str, optional): The working directory. Defaults to ".".
            cores (int, optional): The number of cores. Defaults to 1.
            memory_max (int, optional): The maximum memory. Defaults to None.
            run_time_max (int, optional): The maximum run time. Defaults to None.
            dependency_list (list[int], optional): The list of dependency job IDs. Defaults to None.
            submission_template (str): Submission script template pysqa.wrapper.flux.template

        Returns:
            str: The rendered job submission template.
        """
        return super().render_submission_template(
            command=command,
            job_name=job_name,
            working_directory=working_directory,
            cores=cores,
            memory_max=memory_max,
            run_time_max=run_time_max,
            dependency_list=dependency_list,
            submission_template=submission_template,
            **kwargs,
        )



import os
from abc import ABC, abstractmethod
from typing import List, Optional, Union

import pandas
from jinja2 import Template


class SchedulerCommands(ABC):
    @property
    def enable_reservation_command(self) -> list[str]:
        """
        Returns the command to enable job reservation on the scheduler.

        Returns:
            list[str]: The command to enable job reservation.
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def submit_job_command(self) -> list[str]:
        """
        Returns the command to submit a job to the scheduler.

        Returns:
            list[str]: The command to submit a job.
        """
        pass

    @property
    @abstractmethod
    def delete_job_command(self) -> list[str]:
        """
        Returns the command to delete a job from the scheduler.

        Returns:
            list[str]: The command to delete a job.
        """
        pass

    @property
    @abstractmethod
    def get_queue_status_command(self) -> list[str]:
        """
        Returns the command to get the status of the job queue.

        Returns:
            list[str]: The command to get the queue status.
        """
        pass

    @staticmethod
    def render_submission_template(
        command: str,
        submission_template: Union[str, Template],
        job_name: str = "pysqa",
        working_directory: str = os.path.abspath("."),
        cores: int = 1,
        memory_max: Optional[int] = None,
        run_time_max: Optional[int] = None,
        dependency_list: Optional[List[int]] = None,
        **kwargs,
    ) -> str:
        """
        Generate the job submission template.

        Args:
            command (str, optional): The command to be executed.
            job_name (str, optional): The job name. Defaults to "pysqa".
            working_directory (str, optional): The working directory. Defaults to ".".
            cores (int, optional): The number of cores. Defaults to 1.
            memory_max (int, optional): The maximum memory. Defaults to None.
            run_time_max (int, optional): The maximum run time. Defaults to None.
            dependency_list (list[int], optional): The list of dependency job IDs. Defaults to None.
            submission_template (str): Submission script template pysqa.wrapper.torque.template

        Returns:
            str: The rendered job submission template.
        """
        if not isinstance(submission_template, Template):
            submission_template = Template(submission_template)
        return submission_template.render(
            command=command,
            job_name=job_name,
            working_directory=working_directory,
            cores=cores,
            memory_max=memory_max,
            run_time_max=run_time_max,
            dependency_list=dependency_list,
            **kwargs,
        )

    @staticmethod
    def dependencies(dependency_list: list[str]) -> list:
        """
        Returns the list of dependencies for a job.

        Args:
            dependency_list (list[str]): The list of dependencies.

        Returns:
            list: The list of dependencies.
        """
        if dependency_list is not None:
            raise NotImplementedError()
        else:
            return []

    @staticmethod
    def get_job_id_from_output(queue_submit_output: str) -> int:
        """
        Returns the job ID from the output of the job submission command.

        Args:
            queue_submit_output (str): The output of the job submission command.

        Returns:
            int: The job ID.
        """
        raise NotImplementedError()

    @staticmethod
    def convert_queue_status(queue_status_output: str) -> pandas.DataFrame:
        """
        Converts the output of the queue status command to a pandas DataFrame.

        Args:
            queue_status_output (str): The output of the queue status command.

        Returns:
            pandas.DataFrame: The queue status as a DataFrame.
        """
        raise NotImplementedError()



import os
import re
from typing import List, Optional, Union

import pandas
from jinja2 import Template

from pysqa.wrapper.abstract import SchedulerCommands

template = """\
#!/bin/bash
#PBS -l ncpus={{cores}}
#PBS -N {{job_name}}
{%- if memory_max %}
#PBS -l mem={{ memory_max| int }}GB
{%- endif %}
{%- if run_time_max %}
#PBS -l walltime={{run_time_max}} 
{%- endif %}
#PBS -l wd
{%- if dependency %}
#PBS -W depend=afterok:{{ dependency | join(':') }}
{%- endif %}
 
{{command}}
"""


class TorqueCommands(SchedulerCommands):
    @property
    def submit_job_command(self) -> list[str]:
        """Returns the command to submit a job."""
        return ["qsub"]

    @property
    def delete_job_command(self) -> list[str]:
        """Returns the command to delete a job."""
        return ["qdel"]

    @property
    def get_queue_status_command(self) -> list[str]:
        """Returns the command to get the queue status."""
        return ["qstat", "-f"]

    @staticmethod
    def get_job_id_from_output(queue_submit_output: str) -> int:
        """Extracts the job ID from the queue submit output.

        Args:
            queue_submit_output (str): The output of the queue submit command.

        Returns:
            int: The job ID.

        Raises:
            ValueError: If the job ID cannot be extracted.
        """
        return int(
            queue_submit_output.splitlines()[-1].rstrip().lstrip().split(sep=".")[0]
        )

    @staticmethod
    def convert_queue_status(queue_status_output: str) -> pandas.DataFrame:
        """Converts the queue status output into a pandas DataFrame.

        Args:
            queue_status_output (str): The output of the queue status command.

        Returns:
            pandas.DataFrame: The queue status as a DataFrame.
        """
        lines = queue_status_output.split("\n")
        input_string = "".join(lines)
        # remove all whitespaces
        input_string = "".join(input_string.split())

        # Extract the job ID, user, job name, status, and working directory for each running job
        regex_pattern_job_id = r"JobId:(.*?)Job_Name"
        job_id_lst = re.findall(regex_pattern_job_id, input_string)
        job_id_lst = [int(job_id_str.split(sep=".")[0]) for job_id_str in job_id_lst]

        regex_pattern_user = r"Job_Owner=(.*?)job_state"
        user_lst = re.findall(regex_pattern_user, input_string)
        user_lst = [user_str.split("@")[0] for user_str in user_lst]

        regex_pattern_job_name = r"Job_Name=(.*?)Job_Owner"
        job_name_lst = re.findall(regex_pattern_job_name, input_string)

        regex_pattern_status = r"job_state=(.*?)queue"
        status_lst = re.findall(regex_pattern_status, input_string)

        regex_pattern_working_directory = r"PBS_O_WORKDIR=(.*?),PBS"
        working_directory_lst = re.findall(
            regex_pattern_working_directory, input_string
        )

        df = pandas.DataFrame(
            {
                "jobid": job_id_lst,
                "user": user_lst,
                "jobname": job_name_lst,
                "status": status_lst,
                "working_directory": working_directory_lst,
            }
        )

        df["status"] = df["status"].apply(
            lambda x: "running" if x == "R" else "pending"
        )

        return df

    def render_submission_template(
        self,
        command: str,
        job_name: str = "pysqa",
        working_directory: str = os.path.abspath("."),
        cores: int = 1,
        memory_max: Optional[int] = None,
        run_time_max: Optional[int] = None,
        dependency_list: Optional[List[int]] = None,
        submission_template: Union[str, Template] = template,
        **kwargs,
    ) -> str:
        """
        Generate the job submission template.

        Args:
            command (str, optional): The command to be executed.
            job_name (str, optional): The job name. Defaults to "pysqa".
            working_directory (str, optional): The working directory. Defaults to ".".
            cores (int, optional): The number of cores. Defaults to 1.
            memory_max (int, optional): The maximum memory. Defaults to None.
            run_time_max (int, optional): The maximum run time. Defaults to None.
            dependency_list (list[int], optional): The list of dependency job IDs. Defaults to None.
            submission_template (str): Submission script template pysqa.wrapper.flux.template

        Returns:
            str: The rendered job submission template.
        """
        return super().render_submission_template(
            command=command,
            job_name=job_name,
            working_directory=working_directory,
            cores=cores,
            memory_max=memory_max,
            run_time_max=run_time_max,
            dependency_list=dependency_list,
            submission_template=submission_template,
            **kwargs,
        )



import os
import unittest
from pysqa import QueueAdapter

try:
    import paramiko
    from tqdm import tqdm

    skip_multi_test = False
except ImportError:
    skip_multi_test = True


@unittest.skipIf(
    skip_multi_test,
    "Either paramiko or tqdm are not installed, so the multi queue adapter tests are skipped.",
)
class TestMultiQueueAdapter(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.path = os.path.dirname(os.path.abspath(__file__))
        cls.multi = QueueAdapter(
            directory=os.path.join(cls.path, "config/multicluster")
        )

    def test_config(self):
        self.assertEqual(self.multi.config["queue_type"], "SLURM")
        self.assertEqual(self.multi.config["queue_primary"], "slurm")

    def test_list_clusters(self):
        self.assertEqual(self.multi.list_clusters(), ["local_slurm", "remote_slurm"])



import os
from typing import List, Optional, Union

import pandas
from jinja2 import Template

from pysqa.wrapper.abstract import SchedulerCommands

template = """\
#!/bin/bash
#BSUB -q queue
#BSUB -J {{job_name}}
#BSUB -o time.out
#BSUB -n {{cores}}
#BSUB -cwd {{working_directory}}
#BSUB -e error.out
{%- if run_time_max %}
#BSUB -W {{run_time_max}}
{%- endif %}
{%- if memory_max %}
#BSUB -M {{memory_max}}
{%- endif %}

{{command}}
"""


class LsfCommands(SchedulerCommands):
    @property
    def submit_job_command(self) -> list[str]:
        """Return the command to submit a job."""
        return ["bsub"]

    @property
    def delete_job_command(self) -> list[str]:
        """Return the command to delete a job."""
        return ["bkill"]

    @property
    def get_queue_status_command(self) -> list[str]:
        """Return the command to get the queue status."""
        return ["bjobs"]

    @staticmethod
    def get_job_id_from_output(queue_submit_output: str) -> int:
        """Extract the job ID from the queue submit output."""
        return int(queue_submit_output.split("<")[1].split(">")[0])

    @staticmethod
    def convert_queue_status(queue_status_output: str) -> pandas.DataFrame:
        """Convert the queue status output to a pandas DataFrame."""
        job_id_lst, user_lst, status_lst, job_name_lst = [], [], [], []
        line_split_lst = queue_status_output.split("\n")
        if len(line_split_lst) > 1:
            for line in line_split_lst[1:]:
                line_segments = line.split()
                if len(line_segments) > 1:
                    job_id_lst.append(int(line_segments[0]))
                    user_lst.append(line_segments[1])
                    status_lst.append(line_segments[2])
                    job_name_lst.append(line_segments[6])
        df = pandas.DataFrame(
            {
                "jobid": job_id_lst,
                "user": user_lst,
                "jobname": job_name_lst,
                "status": status_lst,
            }
        )
        df.loc[df.status == "RUN", "status"] = "running"
        df.loc[df.status == "PEND", "status"] = "pending"
        return df

    def render_submission_template(
        self,
        command: str,
        job_name: str = "pysqa",
        working_directory: str = os.path.abspath("."),
        cores: int = 1,
        memory_max: Optional[int] = None,
        run_time_max: Optional[int] = None,
        dependency_list: Optional[List[int]] = None,
        submission_template: Union[str, Template] = template,
        **kwargs,
    ) -> str:
        """
        Generate the job submission template.

        Args:
            command (str, optional): The command to be executed.
            job_name (str, optional): The job name. Defaults to "pysqa".
            working_directory (str, optional): The working directory. Defaults to ".".
            cores (int, optional): The number of cores. Defaults to 1.
            memory_max (int, optional): The maximum memory. Defaults to None.
            run_time_max (int, optional): The maximum run time. Defaults to None.
            dependency_list (list[int], optional): The list of dependency job IDs. Defaults to None.
            submission_template (str): Submission script template pysqa.wrapper.flux.template

        Returns:
            str: The rendered job submission template.
        """
        return super().render_submission_template(
            command=command,
            job_name=job_name,
            working_directory=working_directory,
            cores=cores,
            memory_max=memory_max,
            run_time_max=run_time_max,
            dependency_list=dependency_list,
            submission_template=submission_template,
            **kwargs,
        )



from typing import Union

import pandas

from pysqa.wrapper.slurm import SlurmCommands
from pysqa.wrapper.slurm import template as template_slurm

template = template_slurm


class GentCommands(SlurmCommands):
    @staticmethod
    def get_job_id_from_output(queue_submit_output: str) -> int:
        """
        Extracts the job ID from the queue submit output.

        Args:
            queue_submit_output (str): The output of the queue submit command.

        Returns:
            int: The job ID.

        """
        return int(queue_submit_output.splitlines()[-1].rstrip().lstrip().split(";")[0])

    @staticmethod
    def get_queue_from_output(queue_submit_output: str) -> str:
        """
        Extracts the queue name from the queue submit output.

        Args:
            queue_submit_output (str): The output of the queue submit command.

        Returns:
            str: The queue name.

        """
        return str(queue_submit_output.splitlines()[-1].rstrip().lstrip().split(";")[1])

    @staticmethod
    def convert_queue_status(queue_status_output: str) -> Union[pandas.DataFrame, None]:
        """
        Converts the queue status output into a pandas DataFrame.

        Args:
            queue_status_output (str): The output of the queue status command.

        Returns:
            pandas.DataFrame: The converted queue status.

        """
        qstat = queue_status_output.splitlines()
        queue = qstat[0].split(":")[1].strip()
        if len(qstat) <= 1:  # first row contains cluster name, check if there are jobs
            return None

        line_split_lst = [line.split("|") for line in qstat[1:]]
        job_id_lst, user_lst, status_lst, job_name_lst, queue_lst = zip(
            *[
                (int(jobid), user, status.lower(), jobname, queue)
                for jobid, user, status, jobname in line_split_lst
            ]
        )
        return pandas.DataFrame(
            {
                "cluster": queue_lst,
                "jobid": job_id_lst,
                "user": user_lst,
                "jobname": job_name_lst,
                "status": status_lst,
            }
        )

    @staticmethod
    def dependencies(dependency_list: list[str]) -> list[str]:
        """
        Returns the dependencies for the job.

        Args:
            dependency_list (list[str]): The list of job dependencies.

        Returns:
            list[str]: The dependencies for the job.

        Raises:
            NotImplementedError: If dependency_list is not None.

        """
        if dependency_list is not None:
            raise NotImplementedError()
        else:
            return []



import getpass
import importlib
import os
import subprocess
from typing import List, Optional, Tuple, Union

import pandas
from jinja2 import Template

from pysqa.base.abstract import QueueAdapterAbstractClass
from pysqa.wrapper.abstract import SchedulerCommands

queue_type_dict = {
    "SGE": {
        "class_name": "SunGridEngineCommands",
        "module_name": "pysqa.wrapper.sge",
    },
    "TORQUE": {
        "class_name": "TorqueCommands",
        "module_name": "pysqa.wrapper.torque",
    },
    "SLURM": {
        "class_name": "SlurmCommands",
        "module_name": "pysqa.wrapper.slurm",
    },
    "LSF": {
        "class_name": "LsfCommands",
        "module_name": "pysqa.wrapper.lsf",
    },
    "MOAB": {
        "class_name": "MoabCommands",
        "module_name": "pysqa.wrapper.moab",
    },
    "GENT": {
        "class_name": "GentCommands",
        "module_name": "pysqa.wrapper.gent",
    },
    "REMOTE": {
        "class_name": None,
        "module_name": None,
    },
    "FLUX": {
        "class_name": "FluxCommands",
        "module_name": "pysqa.wrapper.flux",
    },
}


def execute_command(
    commands: str,
    working_directory: Optional[str] = None,
    split_output: bool = True,
    shell: bool = False,
    error_filename: str = "pysqa.err",
) -> Union[str, List[str]]:
    """
    A wrapper around the subprocess.check_output function.

    Args:
        commands (str): The command(s) to be executed on the command line
        working_directory (str, optional): The directory where the command is executed. Defaults to None.
        split_output (bool, optional): Boolean flag to split newlines in the output. Defaults to True.
        shell (bool, optional): Additional switch to convert commands to a single string. Defaults to False.
        error_filename (str, optional): In case the execution fails, the output is written to this file. Defaults to "pysqa.err".

    Returns:
        Union[str, List[str]]: Output of the shell command either as a string or as a list of strings
    """
    if shell and isinstance(commands, list):
        commands = " ".join(commands)
    try:
        out = subprocess.check_output(
            commands,
            cwd=working_directory,
            stderr=subprocess.STDOUT,
            universal_newlines=True,
            shell=not isinstance(commands, list),
        )
    except subprocess.CalledProcessError as e:
        with open(os.path.join(working_directory, error_filename), "w") as f:
            print(e.stdout, file=f)
        out = None
    if out is not None and split_output:
        return out.split("\n")
    else:
        return out


def get_queue_commands(queue_type: str) -> Union[SchedulerCommands, None]:
    """
    Load queuing system commands class

    Args:
        queue_type (str): Type of the queuing system in capital letters

    Returns:
        SchedulerCommands: queuing system commands class instance
    """
    if queue_type in queue_type_dict.keys():
        class_name = queue_type_dict[queue_type]["class_name"]
        module_name = queue_type_dict[queue_type]["module_name"]
        if module_name is not None and class_name is not None:
            return getattr(importlib.import_module(module_name), class_name)()
        else:
            return None
    else:
        raise ValueError(
            "The queue_type "
            + queue_type
            + " is not found in the list of supported queue types "
            + str(list(queue_type_dict.keys()))
        )


class QueueAdapterCore(QueueAdapterAbstractClass):
    """
    The goal of the QueueAdapter class is to make submitting to a queue system as easy as starting another sub process
    locally.

    Args:
        queue_type (str): Type of the queuing system in capital letters
        execute_command (funct): Function to execute commands.
    """

    def __init__(
        self,
        queue_type: str,
        execute_command: callable = execute_command,
    ):
        self._commands = get_queue_commands(queue_type=queue_type)
        if queue_type_dict[queue_type]["module_name"] is not None:
            self._submission_template = getattr(
                importlib.import_module(queue_type_dict[queue_type]["module_name"]),
                "template",
            )
        self._execute_command_function = execute_command

    def submit_job(
        self,
        queue: Optional[str] = None,
        job_name: Optional[str] = None,
        working_directory: Optional[str] = None,
        cores: Optional[int] = None,
        memory_max: Optional[int] = None,
        run_time_max: Optional[int] = None,
        dependency_list: Optional[List[str]] = None,
        command: Optional[str] = None,
        submission_template: Optional[Union[str, Template]] = None,
        **kwargs,
    ) -> Union[int, None]:
        """
        Submit a job to the queue.

        Args:
            queue (str/None): The queue to submit the job to.
            job_name (str/None): The name of the job.
            working_directory (str/None): The working directory for the job.
            cores (int/None): The number of cores required for the job.
            memory_max (int/None): The maximum memory required for the job.
            run_time_max (int/None): The maximum run time for the job.
            dependency_list (list[str]/None): List of job dependencies.
            command (str/None): The command to execute for the job.

        Returns:
            int: The job ID.
        """
        if working_directory is not None and " " in working_directory:
            raise ValueError(
                "Whitespaces in the working_directory name are not supported!"
            )
        if submission_template is None:
            submission_template = self._submission_template
        working_directory, queue_script_path = self._write_queue_script(
            queue=queue,
            job_name=job_name,
            working_directory=working_directory,
            cores=cores,
            memory_max=memory_max,
            run_time_max=run_time_max,
            command=command,
            dependency_list=dependency_list,
            submission_template=submission_template,
            **kwargs,
        )
        out = self._execute_command(
            commands=self._list_command_to_be_executed(
                queue_script_path=queue_script_path
            ),
            working_directory=working_directory,
            split_output=False,
        )
        if out is not None:
            return self._commands.get_job_id_from_output(out)
        else:
            return None

    def enable_reservation(self, process_id: int):
        """
        Enable reservation for a process.

        Args:
            process_id (int): The process ID.

        Returns:
            str: The result of the enable reservation command.
        """
        out = self._execute_command(
            commands=self._commands.enable_reservation_command + [str(process_id)],
            split_output=True,
        )
        if out is not None:
            return out[0]
        else:
            return None

    def delete_job(self, process_id: int) -> Union[str, None]:
        """
        Delete a job.

        Args:
            process_id (int): The process ID.

        Returns:
            str: The result of the delete job command.
        """
        out = self._execute_command(
            commands=self._commands.delete_job_command + [str(process_id)],
            split_output=True,
        )
        if out is not None:
            return out[0]
        else:
            return None

    def get_queue_status(self, user: Optional[str] = None) -> pandas.DataFrame:
        """
        Get the status of the queue.

        Args:
            user (str): The user to filter the queue status for.

        Returns:
            pandas.DataFrame: The queue status.
        """
        out = self._execute_command(
            commands=self._commands.get_queue_status_command, split_output=False
        )
        df = self._commands.convert_queue_status(queue_status_output=out)
        if user is None:
            return df
        else:
            return df[df["user"] == user]

    def get_status_of_my_jobs(self) -> pandas.DataFrame:
        """
        Get the status of the user's jobs.

        Returns:
            pandas.DataFrame: The status of the user's jobs.
        """
        return self.get_queue_status(user=self._get_user())

    def get_status_of_job(self, process_id: int) -> Union[str, None]:
        """
        Get the status of a job.

        Args:
            process_id (int): The process ID.

        Returns:
            str: The status of the job.results_lst.append(df_selected.values[0])
        """
        df = self.get_queue_status()
        df_selected = df[df["jobid"] == process_id]["status"]
        if len(df_selected) != 0:
            return df_selected.values[0]
        else:
            return None

    def get_status_of_jobs(self, process_id_lst: List[int]) -> List[str]:
        """
        Get the status of multiple jobs.

        Args:
            process_id_lst (list[int]): List of process IDs.

        Returns:
            list[str]: List of job statuses.
        """
        df = self.get_queue_status()
        results_lst = []
        for process_id in process_id_lst:
            df_selected = df[df["jobid"] == process_id]["status"]
            if len(df_selected) != 0:
                results_lst.append(df_selected.values[0])
            else:
                results_lst.append("finished")
        return results_lst

    def _list_command_to_be_executed(self, queue_script_path: str) -> list:
        """
        Get the list of commands to be executed.

        Args:
            queue_script_path (str): The path to the queue script.

        Returns:
            list: The list of commands to be executed.
        """
        return self._commands.submit_job_command + [queue_script_path]

    def _execute_command(
        self,
        commands: Union[str, List[str]],
        working_directory: Optional[str] = None,
        split_output: bool = True,
        shell: bool = False,
        error_filename: str = "pysqa.err",
    ) -> str:
        """
        Execute a command or a list of commands.

        Args:
            commands (Union[str, List[str]]): The command(s) to be executed.
            working_directory (Optional[str], optional): The working directory. Defaults to None.
            split_output (bool, optional): Whether to split the output into lines. Defaults to True.
            shell (bool, optional): Whether to use the shell to execute the command. Defaults to False.
            error_filename (str, optional): The name of the error file. Defaults to "pysqa.err".

        Returns:
            str: The output of the command(s).
        """
        return self._execute_command_function(
            commands=commands,
            working_directory=working_directory,
            split_output=split_output,
            shell=shell,
            error_filename=error_filename,
        )

    def _write_queue_script(
        self,
        queue: Optional[str] = None,
        submission_template: Optional[Union[str, Template]] = None,
        job_name: Optional[str] = None,
        working_directory: Optional[str] = None,
        cores: Optional[int] = None,
        memory_max: Optional[int] = None,
        run_time_max: Optional[int] = None,
        dependency_list: Optional[List[int]] = None,
        command: Optional[str] = None,
        **kwargs,
    ) -> Tuple[str, str]:
        """
        Write the queue script to a file.

        Args:
            queue (str/None): The queue name.
            job_name (str/None): The job name.
            working_directory (str/None): The working directory.
            cores (int/None): The number of cores.
            memory_max (int/None): The maximum memory.
            run_time_max (int/None): The maximum run time.
            dependency_list (list/None): The list of dependency job IDs.
            command (str/None): The command to be executed.

        Returns:
            Tuple[str, str]: A tuple containing the working directory and the path to the queue script file.
        """
        if isinstance(command, list):
            command = "".join(command)
        if working_directory is None:
            working_directory = "."
        queue_script = self._job_submission_template(
            queue=queue,
            submission_template=submission_template,
            job_name=job_name,
            working_directory=working_directory,
            cores=cores,
            memory_max=memory_max,
            run_time_max=run_time_max,
            dependency_list=dependency_list,
            command=command,
            **kwargs,
        )
        if not os.path.exists(working_directory):
            os.makedirs(working_directory)
        queue_script_path = os.path.join(working_directory, "run_queue.sh")
        with open(queue_script_path, "w") as f:
            f.writelines(queue_script)
        return working_directory, queue_script_path

    def _job_submission_template(
        self,
        queue: Optional[str] = None,
        submission_template: Optional[Union[str, Template]] = None,
        job_name: str = "job.py",
        working_directory: str = ".",
        cores: Optional[int] = None,
        memory_max: Optional[int] = None,
        run_time_max: Optional[int] = None,
        dependency_list: Optional[List[int]] = None,
        command: Optional[str] = None,
        **kwargs,
    ) -> str:
        """
        Generate the job submission template.

        Args:
            queue (str, optional): The queue name. Defaults to None.
            job_name (str, optional): The job name. Defaults to "job.py".
            working_directory (str, optional): The working directory. Defaults to ".".
            cores (int, optional): The number of cores. Defaults to None.
            memory_max (int, optional): The maximum memory. Defaults to None.
            run_time_max (int, optional): The maximum run time. Defaults to None.
            dependency_list (list[int], optional): The list of dependency job IDs. Defaults to None.
            command (str, optional): The command to be executed. Defaults to None.

        Returns:
            str: The job submission template.
        """
        if queue is not None:
            raise ValueError()
        if submission_template is None:
            submission_template = self._submission_template
        return self._commands.render_submission_template(
            command=command,
            submission_template=submission_template,
            working_directory=working_directory,
            job_name=job_name,
            cores=cores,
            memory_max=memory_max,
            run_time_max=run_time_max,
            dependency_list=dependency_list,
            **kwargs,
        )

    @staticmethod
    def _get_user() -> str:
        """
        Get the current user.

        Returns:
            str: The current user.
        """
        return getpass.getuser()



import os
from typing import List, Optional, Tuple, Union

import pandas
import yaml
from jinja2 import Template
from jinja2.exceptions import TemplateSyntaxError

from pysqa.base.core import QueueAdapterCore, execute_command
from pysqa.base.validate import check_queue_parameters, value_error_if_none


class Queues(object):
    """
    Queues is an abstract class simply to make the list of queues available for auto completion. This is mainly used in
    interactive environments like jupyter.
    """

    def __init__(self, list_of_queues: List[str]):
        """
        Initialize the Queues object.

        Args:
            list_of_queues (List[str]): A list of queue names.

        """
        self._list_of_queues = list_of_queues

    def __getattr__(self, item: str) -> str:
        """
        Get the queue name.

        Args:
            item (str): The name of the queue.

        Returns:
            str: The name of the queue.

        Raises:
            AttributeError: If the queue name is not in the list of queues.

        """
        if item in self._list_of_queues:
            return item
        else:
            raise AttributeError

    def __dir__(self) -> List[str]:
        """
        Get the list of queues.

        Returns:
            List[str]: The list of queues.

        """
        return self._list_of_queues


class QueueAdapterWithConfig(QueueAdapterCore):
    """
    The goal of the QueueAdapter class is to make submitting to a queue system as easy as starting another sub process
    locally.

    Args:
        config (dict): Configuration for the QueueAdapter.
        directory (str): Directory containing the queue.yaml files as well as corresponding jinja2 templates for the individual queues.
        execute_command(funct): Function to execute commands.
    """

    def __init__(
        self,
        config: dict,
        directory: str = "~/.queues",
        execute_command: callable = execute_command,
    ):
        super().__init__(
            queue_type=config["queue_type"], execute_command=execute_command
        )
        self._config = config
        self._fill_queue_dict(queue_lst_dict=self._config["queues"])
        self._load_templates(queue_lst_dict=self._config["queues"], directory=directory)
        self._queues = Queues(self.queue_list)
        self._remote_flag = False
        self._ssh_delete_file_on_remote = True

    @property
    def ssh_delete_file_on_remote(self) -> bool:
        """
        Get the value of ssh_delete_file_on_remote.

        Returns:
            bool: The value of ssh_delete_file_on_remote.
        """
        return self._ssh_delete_file_on_remote

    @property
    def remote_flag(self) -> bool:
        """
        Get the value of remote_flag.

        Returns:
            bool: The value of remote_flag.
        """
        return self._remote_flag

    @property
    def config(self) -> dict:
        """
        Get the QueueAdapter configuration.

        Returns:
            dict: The QueueAdapter configuration.
        """
        return self._config

    @property
    def queue_list(self) -> list:
        """
        Get the list of available queues.

        Returns:
            list: The list of available queues.
        """
        return list(self._config["queues"].keys())

    @property
    def queue_view(self) -> pandas.DataFrame:
        """
        Get the Pandas DataFrame representation of the available queues.

        Returns:
            pandas.DataFrame: The Pandas DataFrame representation of the available queues.
        """
        return pandas.DataFrame(self._config["queues"]).T.drop(
            ["script", "template"], axis=1
        )

    @property
    def queues(self):
        """
        Get the available queues.

        Returns:
            Queues: The available queues.
        """
        return self._queues

    def get_job_from_remote(self, working_directory: str) -> None:
        """
        Get the results of the calculation - this is necessary when the calculation was executed on a remote host.

        Args:
            working_directory (str): The working directory where the calculation was executed.
        """
        raise NotImplementedError

    def convert_path_to_remote(self, path: str):
        """
        Converts a local file path to a remote file path.

        Args:
            path (str): The local file path to be converted.

        Returns:
            str: The converted remote file path.
        """
        raise NotImplementedError

    def transfer_file(
        self,
        file: str,
        transfer_back: bool = False,
        delete_file_on_remote: bool = False,
    ):
        """
        Transfer a file to a remote location.

        Args:
            file (str): The path of the file to be transferred.
            transfer_back (bool, optional): Whether to transfer the file back after processing. Defaults to False.
            delete_file_on_remote (bool, optional): Whether to delete the file on the remote location after transfer. Defaults to False.
        """
        raise NotImplementedError

    def check_queue_parameters(
        self,
        queue: Optional[str],
        cores: int = 1,
        run_time_max: Optional[int] = None,
        memory_max: Optional[int] = None,
        active_queue: Optional[dict] = None,
    ) -> Tuple[
        Union[float, int, None], Union[float, int, None], Union[float, int, None]
    ]:
        """
        Check the parameters of a queue.

        Args:
            queue (str): The queue to check.
            cores (int, optional): The number of cores. Defaults to 1.
            run_time_max (int, optional): The maximum run time. Defaults to None.
            memory_max (int, optional): The maximum memory. Defaults to None.
            active_queue (dict, optional): The active queue. Defaults to None.

        Returns:
            list: A list of queue parameters [cores, run_time_max, memory_max].
        """
        if active_queue is None:
            active_queue = self._config["queues"][queue]
        return check_queue_parameters(
            active_queue=active_queue,
            cores=cores,
            run_time_max=run_time_max,
            memory_max=memory_max,
        )

    def _job_submission_template(
        self,
        queue: Optional[str] = None,
        submission_template: Optional[Union[str, Template]] = None,
        job_name: str = "job.py",
        working_directory: str = ".",
        cores: Optional[int] = None,
        memory_max: Optional[int] = None,
        run_time_max: Optional[int] = None,
        dependency_list: Optional[List[int]] = None,
        command: Optional[str] = None,
        **kwargs,
    ) -> str:
        """
        Generate the job submission template.

        Args:
            queue (str, optional): The queue name. Defaults to None.
            job_name (str, optional): The job name. Defaults to "job.py".
            working_directory (str, optional): The working directory. Defaults to ".".
            cores (int, optional): The number of cores. Defaults to None.
            memory_max (int, optional): The maximum memory. Defaults to None.
            run_time_max (int, optional): The maximum run time. Defaults to None.
            dependency_list (list[int], optional): The list of dependency job IDs. Defaults to None.
            command (str, optional): The command to be executed. Defaults to None.

        Returns:
            str: The job submission template.
        """
        if queue is None:
            queue = self._config["queue_primary"]
        value_error_if_none(value=command)
        if queue not in self.queue_list:
            raise ValueError(
                "The queue "
                + queue
                + " was not found in the list of queues: "
                + str(self.queue_list)
            )
        active_queue = self._config["queues"][queue]
        cores, run_time_max, memory_max = self.check_queue_parameters(
            queue=None,
            cores=cores,
            run_time_max=run_time_max,
            memory_max=memory_max,
            active_queue=active_queue,
        )
        return super()._job_submission_template(
            queue=None,
            submission_template=active_queue["template"],
            job_name=job_name,
            working_directory=working_directory,
            cores=cores,
            memory_max=memory_max,
            run_time_max=run_time_max,
            dependency_list=dependency_list,
            command=command,
            **kwargs,
        )

    @staticmethod
    def _fill_queue_dict(queue_lst_dict: dict):
        """
        Fill missing keys in the queue dictionary with None values.

        Args:
            queue_lst_dict (dict): The queue dictionary.
        """
        queue_keys = ["cores_min", "cores_max", "run_time_max", "memory_max"]
        for queue_dict in queue_lst_dict.values():
            for key in set(queue_keys) - set(queue_dict.keys()):
                queue_dict[key] = None

    @staticmethod
    def _load_templates(queue_lst_dict: dict, directory: str = ".") -> None:
        """
        Load the queue templates from files and store them in the queue dictionary.

        Args:
            queue_lst_dict (dict): The queue dictionary.
            directory (str, optional): The directory where the queue template files are located. Defaults to ".".
        """
        for queue_dict in queue_lst_dict.values():
            if "script" in queue_dict.keys():
                with open(os.path.join(directory, queue_dict["script"]), "r") as f:
                    try:
                        queue_dict["template"] = Template(f.read())
                    except TemplateSyntaxError as error:
                        raise TemplateSyntaxError(
                            message="File: "
                            + queue_dict["script"]
                            + " - "
                            + error.message,
                            lineno=error.lineno,
                        )


def read_config(file_name: str = "queue.yaml") -> dict:
    """
    Read and parse a YAML configuration file.

    Args:
        file_name (str): The name of the YAML file to read.

    Returns:
        dict: The parsed configuration as a dictionary.
    """
    with open(file_name, "r") as f:
        return yaml.load(f, Loader=yaml.FullLoader)



import os
import io
import json
import unittest.mock
from pysqa.base.cmd import command_line


class TestCMD(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.test_dir = os.path.abspath(os.path.dirname(__file__))

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def assert_stdout_command_line(
        self, cmd_args, execute_command, expected_output, mock_stdout
    ):
        command_line(arguments_lst=cmd_args, execute_command=execute_command)
        self.assertEqual(mock_stdout.getvalue(), expected_output)

    def test_help(self):
        self.assert_stdout_command_line(
            ["--help"],
            None,
            "python -m pysqa --help ... coming soon.\n",
        )

    def test_wrong_option(self):
        self.assert_stdout_command_line(
            ["--error"],
            None,
            "python -m pysqa --help\n",
        )

    def test_submit(self):
        def execute_command(
            commands,
            working_directory=None,
            split_output=True,
            shell=False,
            error_filename="pysqa.err",
        ):
            return "1\n"

        self.assert_stdout_command_line(
            [
                "--config_directory",
                os.path.join(self.test_dir, "config", "slurm"),
                "--submit",
                "--queue",
                "slurm",
                "--job_name",
                "test",
                "--working_directory",
                ".",
                "--cores",
                "2",
                "--memory",
                "1GB",
                "--run_time",
                "10",
                "--command",
                "echo hello",
            ],
            execute_command,
            "1\n",
        )
        with open("run_queue.sh") as f:
            output = f.readlines()
        content = [
            "#!/bin/bash\n",
            "#SBATCH --output=time.out\n",
            "#SBATCH --job-name=test\n",
            "#SBATCH --chdir=.\n",
            "#SBATCH --get-user-env=L\n",
            "#SBATCH --partition=slurm\n",
            "#SBATCH --time=4320\n",
            "#SBATCH --mem=1GBG\n",
            "#SBATCH --cpus-per-task=10\n",
            "\n",
            "echo hello",
        ]
        self.assertEqual(output, content)
        os.remove("run_queue.sh")

    def test_delete(self):
        def execute_command(
            commands,
            working_directory=None,
            split_output=True,
            shell=False,
            error_filename="pysqa.err",
        ):
            return "Success\n"

        self.assert_stdout_command_line(
            [
                "--config_directory",
                os.path.join(self.test_dir, "config", "slurm"),
                "--delete",
                "--id",
                "1",
            ],
            execute_command,
            "S\n",
        )

    def test_status(self):
        def execute_command(
            commands,
            working_directory=None,
            split_output=True,
            shell=False,
            error_filename="pysqa.err",
        ):
            with open(
                os.path.join(self.test_dir, "config", "slurm", "squeue_output")
            ) as f:
                return f.read()

        self.assert_stdout_command_line(
            [
                "--config_directory",
                os.path.join(self.test_dir, "config", "slurm"),
                "--status",
            ],
            execute_command,
            json.dumps(
                {
                    "jobid": [5322019, 5322016, 5322017, 5322018, 5322013],
                    "user": ["janj", "janj", "janj", "janj", "maxi"],
                    "jobname": [
                        "pi_19576488",
                        "pi_19576485",
                        "pi_19576486",
                        "pi_19576487",
                        "pi_19576482",
                    ],
                    "status": ["running", "running", "running", "running", "running"],
                    "working_directory": [
                        "/cmmc/u/janj/pyiron/projects/2023/2023-04-19-dft-test/job_1",
                        "/cmmc/u/janj/pyiron/projects/2023/2023-04-19-dft-test/job_2",
                        "/cmmc/u/janj/pyiron/projects/2023/2023-04-19-dft-test/job_3",
                        "/cmmc/u/janj/pyiron/projects/2023/2023-04-19-dft-test/job_4",
                        "/cmmc/u/janj/pyiron/projects/2023/2023-04-19-dft-test/job_5",
                    ],
                }
            )
            + "\n",
        )

    def test_list(self):
        def execute_command(
            commands,
            working_directory=None,
            split_output=True,
            shell=False,
            error_filename="pysqa.err",
        ):
            pass

        self.assert_stdout_command_line(
            [
                "--config_directory",
                os.path.join(self.test_dir, "config", "slurm"),
                "--list",
                "--working_directory",
                os.path.join(self.test_dir, "config", "slurm"),
            ],
            execute_command,
            json.dumps(
                {
                    "dirs": [os.path.join(self.test_dir, "config", "slurm")],
                    "files": sorted(
                        [
                            os.path.join(
                                self.test_dir, "config", "slurm", "squeue_output"
                            ),
                            os.path.join(
                                self.test_dir, "config", "slurm", "slurm_extra.sh"
                            ),
                            os.path.join(self.test_dir, "config", "slurm", "slurm.sh"),
                            os.path.join(
                                self.test_dir, "config", "slurm", "queue.yaml"
                            ),
                        ]
                    ),
                }
            )
            + "\n",
        )



import os
from typing import List, Optional, Union

import pandas
from flux.job import JobID
from jinja2 import Template

from pysqa.wrapper.abstract import SchedulerCommands

template = """\
#!/bin/bash
# flux: --job-name={{job_name}}
# flux: --env=CORES={{cores}}
# flux: --output=time.out
# flux: --error=error.out
# flux: -n {{cores}}
{%- if run_time_max %}
# flux: -t {{ [1, run_time_max // 60]|max }}
{%- endif %}
{%- if dependency %}
{%- for jobid in dependency %}
# flux: --dependency=afterok:{{jobid}}
{%- endfor %}
{%- endif %}

{{command}}
"""


class FluxCommands(SchedulerCommands):
    @property
    def submit_job_command(self) -> list[str]:
        """Returns the command to submit a job."""
        return ["flux", "batch"]

    @property
    def delete_job_command(self) -> list[str]:
        """Returns the command to delete a job."""
        return ["flux", "cancel"]

    @property
    def get_queue_status_command(self) -> list[str]:
        """Returns the command to get the queue status."""
        return ["flux", "jobs", "-a", "--no-header"]

    @staticmethod
    def get_job_id_from_output(queue_submit_output: str) -> int:
        """Extracts the job ID from the output of the queue submit command."""
        return int(
            JobID(queue_submit_output.splitlines()[-1].rstrip().lstrip().split()[-1])
        )

    @staticmethod
    def convert_queue_status(queue_status_output: str) -> pandas.DataFrame:
        """Converts the queue status output into a pandas DataFrame."""
        line_split_lst = [line.split() for line in queue_status_output.splitlines()]
        job_id_lst, user_lst, job_name_lst, status_lst = [], [], [], []
        for entry in line_split_lst:
            job_id_lst.append(JobID(entry[0]))
            user_lst.append(entry[1])
            job_name_lst.append(entry[2])
            status_lst.append(entry[3])
        df = pandas.DataFrame(
            {
                "jobid": job_id_lst,
                "user": user_lst,
                "jobname": job_name_lst,
                "status": status_lst,
            }
        )
        df.loc[df.status == "R", "status"] = "running"
        df.loc[df.status == "S", "status"] = "pending"
        df.loc[df.status == "C", "status"] = "error"
        df.loc[df.status == "CD", "status"] = "finished"
        return df

    def render_submission_template(
        self,
        command: str,
        job_name: str = "pysqa",
        working_directory: str = os.path.abspath("."),
        cores: int = 1,
        memory_max: Optional[int] = None,
        run_time_max: Optional[int] = None,
        dependency_list: Optional[List[int]] = None,
        submission_template: Union[str, Template] = template,
        **kwargs,
    ) -> str:
        """
        Generate the job submission template.

        Args:
            command (str, optional): The command to be executed.
            job_name (str, optional): The job name. Defaults to "pysqa".
            working_directory (str, optional): The working directory. Defaults to ".".
            cores (int, optional): The number of cores. Defaults to 1.
            memory_max (int, optional): The maximum memory. Defaults to None.
            run_time_max (int, optional): The maximum run time. Defaults to None.
            dependency_list (list[int], optional): The list of dependency job IDs. Defaults to None.
            submission_template (str): Submission script template pysqa.wrapper.flux.template

        Returns:
            str: The rendered job submission template.
        """
        return super().render_submission_template(
            command=command,
            job_name=job_name,
            working_directory=working_directory,
            cores=cores,
            memory_max=memory_max,
            run_time_max=run_time_max,
            dependency_list=dependency_list,
            submission_template=submission_template,
            **kwargs,
        )



from typing import Optional, Union

import pandas

from pysqa.base.config import QueueAdapterWithConfig
from pysqa.base.core import execute_command


class ModularQueueAdapter(QueueAdapterWithConfig):
    """
    A class representing a modular queue adapter.

    Args:
        config (dict): The configuration dictionary.
        directory (str, optional): The directory path. Defaults to "~/.queues".
        execute_command (callable, optional): The execute command function. Defaults to execute_command.

    Attributes:
        _queue_to_cluster_dict (dict): A dictionary mapping queues to clusters.

    Raises:
        ValueError: If a cluster is not found in the list of clusters.

    """

    def __init__(
        self,
        config: dict,
        directory: str = "~/.queues",
        execute_command: callable = execute_command,
    ):
        super(ModularQueueAdapter, self).__init__(
            config=config, directory=directory, execute_command=execute_command
        )
        self._queue_to_cluster_dict = {
            k: v["cluster"] for k, v in self._config["queues"].items()
        }
        for v in self._queue_to_cluster_dict.values():
            if v not in self._config["cluster"]:
                raise ValueError(
                    "The cluster "
                    + v
                    + " was not found in the list of clusters "
                    + str(list(self._config["cluster"].keys()))
                )

    def submit_job(
        self,
        queue: Optional[str] = None,
        job_name: Optional[str] = None,
        working_directory: Optional[str] = None,
        cores: Optional[int] = None,
        memory_max: Optional[str] = None,
        run_time_max: Optional[int] = None,
        dependency_list: Optional[list[str]] = None,
        command: Optional[str] = None,
        **kwargs,
    ) -> Union[int, None]:
        """
        Submit a job to the queue.

        Args:
            queue (str, optional): The queue name. Defaults to None.
            job_name (str, optional): The job name. Defaults to None.
            working_directory (str, optional): The working directory. Defaults to None.
            cores (int, optional): The number of cores. Defaults to None.
            memory_max (int, optional): The maximum memory. Defaults to None.
            run_time_max (int, optional): The maximum run time. Defaults to None.
            dependency_list (list[str], optional): The list of dependencies. Defaults to None.
            command (str, optional): The command to execute. Defaults to None.

        Returns:
            int: The cluster queue ID.

        """
        working_directory, queue_script_path = self._write_queue_script(
            queue=queue,
            job_name=job_name,
            working_directory=working_directory,
            cores=cores,
            memory_max=memory_max,
            run_time_max=run_time_max,
            dependency_list=dependency_list,
            command=command,
            **kwargs,
        )
        cluster_module = self._queue_to_cluster_dict[queue]
        commands = self._switch_cluster_command(
            cluster_module=cluster_module
        ) + self._list_command_to_be_executed(queue_script_path=queue_script_path)
        out = self._execute_command(
            commands=commands,
            working_directory=working_directory,
            split_output=False,
            shell=True,
        )
        if out is not None:
            cluster_queue_id = self._commands.get_job_id_from_output(out)
            cluster_queue_id *= 10
            cluster_queue_id += self._config["cluster"].index(cluster_module)
            return cluster_queue_id
        else:
            return None

    def enable_reservation(self, process_id: int):
        """
        Enable reservation for a process.

        Args:
            process_id (int): The process ID.

        Returns:
            str: The output of the enable reservation command.

        """
        cluster_module, cluster_queue_id = self._resolve_queue_id(
            process_id=process_id, cluster_dict=self._config["cluster"]
        )
        cluster_commands = self._switch_cluster_command(cluster_module=cluster_module)
        commands = (
            cluster_commands
            + self._commands.enable_reservation_command
            + [str(cluster_queue_id)]
        )
        out = self._execute_command(commands=commands, split_output=True, shell=True)
        if out is not None:
            return out[0]
        else:
            return None

    def delete_job(self, process_id: int):
        """
        Delete a job.

        Args:
            process_id (int): The process ID.

        Returns:
            str: The output of the delete job command.

        """
        cluster_module, cluster_queue_id = self._resolve_queue_id(
            process_id=process_id, cluster_dict=self._config["cluster"]
        )
        cluster_commands = self._switch_cluster_command(cluster_module=cluster_module)
        commands = (
            cluster_commands
            + self._commands.delete_job_command
            + [str(cluster_queue_id)]
        )
        out = self._execute_command(commands=commands, split_output=True, shell=True)
        if out is not None:
            return out[0]
        else:
            return None

    def get_queue_status(self, user: Optional[str] = None) -> pandas.DataFrame:
        """
        Get the queue status.

        Args:
            user (str, optional): The user name. Defaults to None.

        Returns:
            pandas.DataFrame: The queue status.

        """
        df_lst = []
        for cluster_module in self._config["cluster"]:
            cluster_commands = self._switch_cluster_command(
                cluster_module=cluster_module
            )
            out = self._execute_command(
                commands=cluster_commands + self._commands.get_queue_status_command,
                split_output=False,
                shell=True,
            )
            df = self._commands.convert_queue_status(queue_status_output=out)
            df_lst.append(df)
        df = pandas.concat(df_lst, axis=0, sort=False).reset_index(drop=True)
        if user is None:
            return df
        else:
            return df[df["user"] == user]

    @staticmethod
    def _resolve_queue_id(process_id: int, cluster_dict: dict):
        """
        Resolve the queue ID.

        Args:
            process_id (int): The process ID.
            cluster_dict (dict): The cluster dictionary.

        Returns:
            tuple: The cluster module and cluster queue ID.

        """
        cluster_queue_id = int(process_id / 10)
        cluster_module = cluster_dict[process_id - cluster_queue_id * 10]
        return cluster_module, cluster_queue_id

    @staticmethod
    def _switch_cluster_command(cluster_module: str):
        """
        Generate the switch cluster command.

        Args:
            cluster_module (str): The cluster module.

        Returns:
            list: The switch cluster command.

        """
        return ["module", "--quiet", "swap", "cluster/{};".format(cluster_module)]



import os
from typing import List, Optional, Tuple, Union

import pandas
from jinja2 import Template

from pysqa.base.abstract import QueueAdapterAbstractClass
from pysqa.base.config import QueueAdapterWithConfig, Queues, read_config
from pysqa.base.core import QueueAdapterCore, execute_command
from pysqa.base.modular import ModularQueueAdapter


class QueueAdapter(QueueAdapterAbstractClass):
    """
    The goal of the QueueAdapter class is to make submitting to a queue system as easy as starting another sub process
    locally.

    Args:
        directory (str): directory containing the queue.yaml files as well as corresponding jinja2 templates for the
                         individual queues.

    Attributes:

        .. attribute:: config

            QueueAdapter configuration read from the queue.yaml file.

        .. attribute:: queue_list

            List of available queues

        .. attribute:: queue_view

            Pandas DataFrame representation of the available queues, read from queue.yaml.

        .. attribute:: queues

            Queues available for auto completion QueueAdapter().queues.<queue name> returns the queue name.
    """

    def __init__(
        self,
        directory: Optional[str] = None,
        queue_type: Optional[str] = None,
        execute_command: callable = execute_command,
    ):
        """
        Initialize the QueueAdapter.

        Args:
            directory (str): Directory containing the queue.yaml files and corresponding templates.
            execute_command (callable): Function to execute commands.
        """
        if directory is not None:
            queue_yaml = os.path.join(directory, "queue.yaml")
            clusters_yaml = os.path.join(directory, "clusters.yaml")
            self._adapter = None
            if os.path.exists(queue_yaml):
                self._queue_dict = {
                    "default": set_queue_adapter(
                        config=read_config(file_name=queue_yaml),
                        directory=directory,
                        execute_command=execute_command,
                    )
                }
                primary_queue = "default"
            elif os.path.exists(clusters_yaml):
                config = read_config(file_name=clusters_yaml)
                self._queue_dict = {
                    k: set_queue_adapter(
                        config=read_config(file_name=os.path.join(directory, v)),
                        directory=directory,
                        execute_command=execute_command,
                    )
                    for k, v in config["cluster"].items()
                }
                primary_queue = config["cluster_primary"]
            else:
                raise ValueError(
                    "Neither a queue.yaml file nor a clusters.yaml file were found in "
                    + directory
                )
            self._adapter = self._queue_dict[primary_queue]
        elif queue_type is not None:
            self._queue_dict = {}
            self._adapter = QueueAdapterCore(queue_type=queue_type.upper())
        else:
            raise ValueError()

    def list_clusters(self) -> List[str]:
        """
        List available computing clusters for remote submission

        Returns:
            List of computing clusters
        """
        return list(self._queue_dict.keys())

    def switch_cluster(self, cluster_name: str):
        """
        Switch to a different computing cluster

        Args:
            cluster_name (str): name of the computing cluster
        """
        self._adapter = self._queue_dict[cluster_name]

    @property
    def config(self) -> Union[dict, None]:
        """
        Get the QueueAdapter configuration.

        Returns:
            dict: The QueueAdapter configuration.
        """
        if isinstance(self._adapter, QueueAdapterWithConfig):
            return self._adapter.config
        else:
            return None

    @property
    def ssh_delete_file_on_remote(self) -> bool:
        """
        Get the value of ssh_delete_file_on_remote property.

        Returns:
            bool: The value of ssh_delete_file_on_remote property.
        """
        if isinstance(self._adapter, QueueAdapterWithConfig):
            return self._adapter.ssh_delete_file_on_remote
        else:
            return False

    @property
    def remote_flag(self) -> bool:
        """
        Get the value of remote_flag property.

        Returns:
            bool: The value of remote_flag property.
        """
        if isinstance(self._adapter, QueueAdapterWithConfig):
            return self._adapter.remote_flag
        else:
            return False

    @property
    def queue_list(self) -> Union[List[str], None]:
        """
        Get the list of available queues.

        Returns:
            List[str]: The list of available queues.
        """
        if isinstance(self._adapter, QueueAdapterWithConfig):
            return self._adapter.queue_list
        else:
            return None

    @property
    def queue_view(self) -> Union[pandas.DataFrame, None]:
        """
        Get the Pandas DataFrame representation of the available queues.

        Returns:
            pandas.DataFrame: The Pandas DataFrame representation of the available queues.
        """
        if isinstance(self._adapter, QueueAdapterWithConfig):
            return self._adapter.queue_view
        else:
            return None

    @property
    def queues(self) -> Union[Queues, None]:
        """
        Get the list of available queues.

        Returns:
            List[str]: The list of available queues.
        """
        if isinstance(self._adapter, QueueAdapterWithConfig):
            return self._adapter.queues
        else:
            return None

    def submit_job(
        self,
        queue: Optional[str] = None,
        job_name: Optional[str] = None,
        working_directory: Optional[str] = None,
        cores: Optional[int] = None,
        memory_max: Optional[int] = None,
        run_time_max: Optional[int] = None,
        dependency_list: Optional[List[str]] = None,
        command: Optional[str] = None,
        submission_template: Optional[Union[str, Template]] = None,
        **kwargs,
    ) -> int:
        """
        Submits command to the given queue.

        Args:
            queue (str/None):  Name of the queue to submit to, must be one of the names configured for this adapter
                               (optional)
            job_name (str/None):  Name of the job for the underlying queuing system (optional)
            working_directory (str/None):  Directory to run the job in (optional)
            cores (int/None):  Number of hardware threads requested (optional)
            memory_max (int/None):  Amount of memory requested per node in GB (optional)
            run_time_max (int/None):  Maximum runtime in seconds (optional)
            dependency_list(list[str]/None: Job ids of jobs to be completed before starting (optional)
            command (str/None): shell command to run in the job
            **kwargs: allows writing additional parameters to the job submission script if they are available in the
                      corresponding template.

        Returns:
            int: Job id received from the queuing system for the job which was submitted
        """
        return self._adapter.submit_job(
            queue=queue,
            job_name=job_name,
            working_directory=working_directory,
            cores=cores,
            memory_max=memory_max,
            run_time_max=run_time_max,
            dependency_list=dependency_list,
            command=command,
            submission_template=submission_template,
            **kwargs,
        )

    def enable_reservation(self, process_id: int) -> str:
        """
        Enable reservation for a process.

        Args:
            process_id (int): The process id.

        Returns:
            str: The result of enabling reservation.
        """
        return self._adapter.enable_reservation(process_id=process_id)

    def get_job_from_remote(self, working_directory: str):
        """
        Get the results of the calculation - this is necessary when the calculation was executed on a remote host.

        Args:
            working_directory (str): The working directory.
        """
        if isinstance(self._adapter, QueueAdapterWithConfig):
            self._adapter.get_job_from_remote(working_directory=working_directory)
        else:
            raise TypeError()

    def transfer_file_to_remote(
        self,
        file: str,
        transfer_back: bool = False,
        delete_file_on_remote: bool = False,
    ):
        """
        Transfer file from remote host to local host.

        Args:
            file (str): The file to transfer.
            transfer_back (bool): Whether to transfer the file back.
            delete_file_on_remote (bool): Whether to delete the file on the remote host.
        """
        if isinstance(self._adapter, QueueAdapterWithConfig):
            self._adapter.transfer_file(
                file=file,
                transfer_back=transfer_back,
                delete_file_on_remote=delete_file_on_remote,
            )
        else:
            raise TypeError()

    def convert_path_to_remote(self, path: str) -> str:
        """
        Convert a local path to a remote path.

        Args:
            path (str): The local path.

        Returns:
            str: The remote path.
        """
        if isinstance(self._adapter, QueueAdapterWithConfig):
            return self._adapter.convert_path_to_remote(path=path)
        else:
            raise TypeError()

    def delete_job(self, process_id: int) -> str:
        """
        Delete a job.

        Args:
            process_id (int): The process id.

        Returns:
            str: The result of deleting the job.
        """
        return self._adapter.delete_job(process_id=process_id)

    def get_queue_status(self, user: Optional[str] = None) -> pandas.DataFrame:
        """
        Get the status of the queue.

        Args:
            user (str/None): The user.

        Returns:
            pandas.DataFrame: The status of the queue.
        """
        return self._adapter.get_queue_status(user=user)

    def get_status_of_my_jobs(self) -> pandas.DataFrame:
        """
        Get the status of the user's jobs.

        Returns:
           pandas.DataFrame: The status of the user's jobs.
        """
        return self._adapter.get_status_of_my_jobs()

    def get_status_of_job(self, process_id: int) -> str:
        """
        Get the status of a job.

        Args:
            process_id: The process id.

        Returns:
             str: The status of the job. Possible values are ['running', 'pending', 'error'].
        """
        return self._adapter.get_status_of_job(process_id=process_id)

    def get_status_of_jobs(self, process_id_lst: List[int]) -> List[str]:
        """
        Get the status of multiple jobs.

        Args:
            process_id_lst: The list of process ids.

        Returns:
             List[str]: The status of the jobs. Possible values are ['running', 'pending', 'error', ...].
        """
        return self._adapter.get_status_of_jobs(process_id_lst=process_id_lst)

    def check_queue_parameters(
        self,
        queue: str,
        cores: int = 1,
        run_time_max: Optional[int] = None,
        memory_max: Optional[int] = None,
        active_queue: Optional[dict] = None,
    ) -> Tuple[
        Union[float, int, None], Union[float, int, None], Union[float, int, None]
    ]:
        """
        Check the parameters of a queue.

        Args:
            queue (str/None): The queue name.
            cores (int): The number of cores.
            run_time_max (int/None): The maximum runtime.
            memory_max (int/None): The maximum memory.
            active_queue (dict/None): The active queue.

        Returns:
            List: A list containing the checked parameters [cores, run_time_max, memory_max].
        """
        if isinstance(self._adapter, QueueAdapterWithConfig):
            return self._adapter.check_queue_parameters(
                queue=queue,
                cores=cores,
                run_time_max=run_time_max,
                memory_max=memory_max,
                active_queue=active_queue,
            )
        else:
            return cores, run_time_max, memory_max


def set_queue_adapter(
    config: dict, directory: str, execute_command: callable = execute_command
):
    """
    Initialize the queue adapter

    Args:
        config (dict): configuration for one cluster
        directory (str): directory which contains the queue configurations
    """
    if config["queue_type"] in ["SGE", "TORQUE", "SLURM", "LSF", "MOAB", "FLUX"]:
        return QueueAdapterWithConfig(
            config=config, directory=directory, execute_command=execute_command
        )
    elif config["queue_type"] in ["GENT"]:
        return ModularQueueAdapter(
            config=config, directory=directory, execute_command=execute_command
        )
    elif config["queue_type"] in ["REMOTE"]:
        # The RemoteQueueAdapter has additional dependencies, namely paramiko and tqdm.
        # By moving the import to this line it only fails when the user specifies the
        # RemoteQueueAdapter in their pysqa configuration.
        from pysqa.base.remote import RemoteQueueAdapter

        return RemoteQueueAdapter(
            config=config, directory=directory, execute_command=execute_command
        )
    else:
        raise ValueError(
            "The queue_type "
            + config["queue_type"]
            + " is not found in the list of supported queue types "
            + str(["SGE", "TORQUE", "SLURM", "LSF", "MOAB", "FLUX", "GENT", "REMOTE"])
        )
